@@ -41,6 +41,22 @@ def _contains_quoted_reserved_chars(x):
     return False
 
 
+def _assert_multistatus_success(r):
+    # Xandikos returns a multistatus on PUT.
+    try:
+        root = _parse_xml(r.content)
+    except InvalidXMLResponse:
+        return
+    for status in root.findall('.//{DAV:}status'):
+        parts = status.text.strip().split()
+        try:
+            st = int(parts[1])
+        except (ValueError, IndexError):
+            continue
+        if st < 200 or st >= 400:
+            raise HTTPError('Server error: {}'.format(st))
+
+
 def _normalize_href(base, href):
     '''Normalize the href to be a path only relative to hostname and
     schema.'''
@@ -76,9 +92,25 @@ class InvalidXMLResponse(exceptions.InvalidResponse):
     pass
 
 
+_BAD_XML_CHARS = (
+    b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x0b\x0c\x0e\x0f'
+    b'\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f'
+)
+
+
+def _clean_body(content, bad_chars=_BAD_XML_CHARS):
+    new_content = content.translate(None, bad_chars)
+    if new_content != content:
+        dav_logger.warning(
+            'Your server incorrectly returned ASCII control characters in its '
+            'XML. Vdirsyncer ignores those, but this is a bug in your server.'
+        )
+    return new_content
+
+
 def _parse_xml(content):
     try:
-        return etree.XML(content)
+        return etree.XML(_clean_body(content))
     except etree.ParseError as e:
         raise InvalidXMLResponse('Invalid XML encountered: {}\n'
                                  'Double-check the URLs in your config.'
@@ -134,23 +166,14 @@ class Discover(object):
         _, collection = url.rstrip('/').rsplit('/', 1)
         return urlparse.unquote(collection)
 
-    def find_dav(self):
+    def find_principal(self):
         try:
-            response = self.session.request(
-                'GET', self._well_known_uri, allow_redirects=True,
-                headers=self.session.get_default_headers()
-            )
-            return response.url
+            return self._find_principal_impl('')
         except (HTTPError, exceptions.Error):
-            # The user might not have well-known URLs set up and instead points
-            # vdirsyncer directly to the DAV server.
-            dav_logger.debug('Server does not support well-known URIs.')
-            return ''
-
-    def find_principal(self, url=None):
-        if url is None:
-            url = self.find_dav()
+            dav_logger.debug('Trying out well-known URI')
+            return self._find_principal_impl(self._well_known_uri)
 
+    def _find_principal_impl(self, url):
         headers = self.session.get_default_headers()
         headers['Depth'] = '0'
         body = b"""
@@ -160,8 +183,10 @@ class Discover(object):
             </d:prop>
         </d:propfind>
         """
+
         response = self.session.request('PROPFIND', url, headers=headers,
                                         data=body)
+
         root = _parse_xml(response.content)
         rv = root.find('.//{DAV:}current-user-principal/{DAV:}href')
         if rv is None:
@@ -172,11 +197,10 @@ class Discover(object):
                 'No current-user-principal returned, re-using URL {}'
                 .format(response.url))
             return response.url
-        return urlparse.urljoin(response.url, rv.text)
+        return urlparse.urljoin(response.url, rv.text).rstrip('/') + '/'
 
-    def find_home(self, url=None):
-        if url is None:
-            url = self.find_principal()
+    def find_home(self):
+        url = self.find_principal()
         headers = self.session.get_default_headers()
         headers['Depth'] = '0'
         response = self.session.request('PROPFIND', url,
@@ -188,7 +212,7 @@ class Discover(object):
         rv = root.find('.//' + self._homeset_tag + '/{DAV:}href')
         if rv is None:
             raise InvalidXMLResponse('Couldn\'t find home-set.')
-        return urlparse.urljoin(response.url, rv.text)
+        return urlparse.urljoin(response.url, rv.text).rstrip('/') + '/'
 
     def find_collections(self):
         rv = None
@@ -209,7 +233,7 @@ class Discover(object):
         props = _merge_xml(response.findall(
             '{DAV:}propstat/{DAV:}prop'
         ))
-        if not props:
+        if props is None or not len(props):
             dav_logger.debug('Skipping, missing <prop>: %s', response)
             return False
         if props.find('{DAV:}resourcetype/' + self._resourcetype) \
@@ -507,6 +531,8 @@ class DAVStorage(Storage):
             headers=headers
         )
 
+        _assert_multistatus_success(response)
+
         # The server may not return an etag under certain conditions:
         #
         #   An origin server MUST NOT send a validator header field (Section
@@ -548,8 +574,9 @@ class DAVStorage(Storage):
             headers=headers
         )
 
-    def _parse_prop_responses(self, root):
-        hrefs = set()
+    def _parse_prop_responses(self, root, handled_hrefs=None):
+        if handled_hrefs is None:
+            handled_hrefs = set()
         for response in root.iter('{DAV:}response'):
             href = response.find('{DAV:}href')
             if href is None:
@@ -558,7 +585,7 @@ class DAVStorage(Storage):
 
             href = self._normalize_href(href.text)
 
-            if href in hrefs:
+            if href in handled_hrefs:
                 # Servers that send duplicate hrefs:
                 # - Zimbra
                 #   https://github.com/pimutils/vdirsyncer/issues/88
@@ -569,9 +596,9 @@ class DAVStorage(Storage):
                 continue
 
             props = response.findall('{DAV:}propstat/{DAV:}prop')
-            if not props:
-                dav_logger.warning('Skipping {!r}, properties are missing.'
-                                   .format(href))
+            if props is None or not len(props):
+                dav_logger.debug('Skipping {!r}, properties are missing.'
+                                 .format(href))
                 continue
             else:
                 props = _merge_xml(props)
@@ -582,8 +609,8 @@ class DAVStorage(Storage):
 
             etag = getattr(props.find('{DAV:}getetag'), 'text', '')
             if not etag:
-                dav_logger.warning('Skipping {!r}, etag property is missing.'
-                                   .format(href))
+                dav_logger.debug('Skipping {!r}, etag property is missing.'
+                                 .format(href))
                 continue
 
             contenttype = getattr(props.find('{DAV:}getcontenttype'),
@@ -594,7 +621,7 @@ class DAVStorage(Storage):
                                          self.item_mimetype))
                 continue
 
-            hrefs.add(href)
+            handled_hrefs.add(href)
             yield href, etag, props
 
     def list(self):
@@ -828,12 +855,14 @@ class CalDAVStorage(DAVStorage):
         # the spec which values from WebDAV are actually allowed.
         headers['Depth'] = '1'
 
+        handled_hrefs = set()
+
         for caldavfilter in caldavfilters:
             xml = data.format(caldavfilter=caldavfilter).encode('utf-8')
             response = self.session.request('REPORT', '', data=xml,
                                             headers=headers)
             root = _parse_xml(response.content)
-            rv = self._parse_prop_responses(root)
+            rv = self._parse_prop_responses(root, handled_hrefs)
             for href, etag, _prop in rv:
                 yield href, etag
 
